#!/usr/bin/perl

# HSP 0.2
#
# Author: Trung Nguyen
#
# This program, hsp.pl, is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
# 17:39 26.09.01

#use Benchmark;use vars '$ht0'; $ht0 = new Benchmark;
#END{
#	if (timestr(timediff(new Benchmark,$ht0))=~/(\S+)\s*CPU/i){
#	print "<br><small><b>Required Time: $1 Seconds</b></small>";
#	}
#}

package HSP;

my $VERSION = '0.2';
my %HTSubs;

use subs qw/HSP_ECHO htembed htsendheader htflush htform/;
use vars qw/%form/;
use subs 'exit';
HSP::handle();

sub exit{
	clear_buffer();
	CORE::exit
}

sub handle{

	local %HSP::header = ();
	local $HSP::headersprinted = 0;
	local $HSP::stdout = tie *STDOUT, 'HSP::STDOUT';
	local $HSP::curfile = '';
	local $0 = undef;
	local %ENV = %ENV;
	local %form = ();
	form_read_in();

	(my $path = $ENV{PATH_TRANSLATED})=~tr/\\/\//;
	$path =~s/\/\.\.\///g;
	(my $path_info = $ENV{PATH_INFO})=~tr/\\/\//;

	while(!-f $path && $path ne ''){
		$path =~s/(\/[^\/]*)$// or last;
		$path_info =~s/\Q$1\E$// or raise_error("Cannot read PATH_INFO.");
	}

	$ENV{PATH_TRANSLATED} = $path;
	$ENV{SCRIPT_FILENAME} = $path;
	$ENV{SCRIPT_NAME} = $path_info;
	$0 = $ENV{PATH_TRANSLATED};

	# Print HSPInfo
	if($ENV{QUERY_STRING} =~ /^hspinfo_(\w+)$/i){
		require 'HSPInfo.pm';
		HSPInfo::hspinfo_out($1);
	}

	if(!-f $path){ raise_error("Cannot find filename $ENV{PATH_INFO}\n<a href=$ENV{SCRIPT_NAME}?hspinfo_env>View Environment</a>") }

	my $dir = $HSP::curfile = $path;
	$dir =~s/\/([^\/]*)$//;
	my $file = $1;
	if($file=~/^\_/ || $file=~/(\.inc\b|\binc\.)/i){ raise_error("This file is not for direct access.") }
	chdir $dir or raise_error("Cannot chdir to $dir.");

	my $code = create_code($path);
	if($code){
		if(eval{ $code->(); 1; }){
			clear_buffer();
		}else{
			if($@){ raise_error("$@") }
			else{ raise_error("An error ocurred while executing code.") }
		}
	}else{
		raise_error("No data");
	}
}

sub clear_buffer{
	# Print Headers?
	if(!$HSP::headersprinted){
		$HSP::headersprinted = 1;
		$HSP::header{'content-type'} = 'text/html' if $HSP::header{'content-type'} eq '';
		$HSP::stdout->prepend(
			(map {lc $_ eq 'content-type' ? '' : $_ . ': ' . $HSP::header{$_} . "\n"} keys %HSP::header),
			'Content-Type: ' . $HSP::header{'content-type'} . "\n",
			"\n"
		);
	}
	# Print Buffer
	$HSP::stdout->flush;
}

sub get_mtime{return (stat($_[0]))[9];}

sub create_code{
	my $filename = shift;
	my $symbol = "HSPF::$filename";
	my $mtime = get_mtime($filename);

	if($HTSubs{'HT::'.$symbol} ne ''){
		if($mtime && $mtime == $HTSubs{'MT::'.$symbol}){
			my $sub = $HTSubs{'HT::'.$symbol};
			$HTSubs{'AT::'.$symbol} = time;
			return $sub;
		}
	}

	open(PTDAT001, $filename) or raise_error("Cannot open $filename: $!.");
	my $source = join '', <PTDAT001>;
	$source=~tr/\r//;
	close(PTDAT001) or raise_error($!);

	my $sourcelength = length $source;
	my @tags = (
		['<%', '%>', index($source, '<%')], # ASP Style
		['<?', '?>', index($source, '<?')],
		['<~', '~>', index($source, '<~')],
	);
	if($source=~/<(script)( language\s?=\s?["']?hsp["']?\s?)>(\s*<!--)?/i){
		push @tags, ["<$1$2>$3", "</$1>", $-[0]-1];
	}
	foreach(@tags){
		if($_->[2] < 0){ $_->[2] = $sourcelength }
	}
	@tags = sort {$a->[2] <=> $b->[2] || length $b->[0] <=> length $a->[0]} @tags; #find first or longest
	my $ptag1 = $tags[0]->[0];
	my $ptag2 = $tags[0]->[1];
	my $lptag1 = length $ptag1;
	my $lptag2 = length $ptag2;
	my($start, $end, $code, $pos, $progress, $line);
	$source .= $ptag1.'1'.$ptag2;
        $sourcelength = length $source;
	while(($pos = index($source, $ptag1, $progress)) > -1){
		$start = $pos+$lptag1;
		$end = index($source, $ptag2, $start);
		while(substr($source, $end-1, 1) eq "\\" && $end > -1){
			$end = index($source, $ptag2, $end+1); # Allow escaping code block terminator
		}
                if($end < 0){ raise_error("Cannot find code block terminator at line " . count_lines(substr $source, 0, $pos)) }
		if($pos > $progress){
			$line = substr($source, $progress, $pos-$progress);
			$line=~s/(\\|\')/\\$1/g;
                        if($line=~s/^([^\r\n]*)(\r?\n|\r)//){
				my($first, $eol) = ($1, $2);
				$eol =~s/\n/\\n/g;
				$eol =~s/\r/\\r/g;
				$code .= "HSP_ECHO('$first',\"$eol\");\n";
			}
			$code .= 'HSP_ECHO(\'' . $line . '\');' if $line ne '';
		}
		if($end > $start){
			$line = substr($source, $start, $end-$start);
			#$line =~s/(?:^|\n)[ \t]*\#[^\r\n]*$//s; # Remarks at block end
			$line=~s/^\=/print /;
			$code .= $line.';' if $line ne '';
		}
		$progress = $end+$lptag2;
	}

	my $sub = mksub($code);
	$HTSubs{'HT::'.$symbol} = $sub;
	$HTSubs{'MT::'.$symbol} = $mtime;
	$HTSubs{'AT::'.$symbol} = time;
	return $sub;
}

sub mksub{
	my $sub;
	my $plaincode = "\$sub = sub{$_[0]\n};";
	unless(eval "$plaincode"){
		if($@){ raise_error($@) }
		else{ raise_error("An error ocurred while parsing $HSP::curfile.") }
	}
}

sub raise_error{
	my $e = shift;
	$e =~s/\\"/-QQuote-/g;
	$e =~s/;?HSP_ECHO\(?.*?"/"/s;
	$e =~s/\-QQuote\-/\\"/g;
	$e=~s/\(eval.*?\)/$HSP::curfile/ if $HSP::curfile ne '';
	print "<p><tt>HSP Error: $e";
	print $HSP::curfile ? "<br>File: $HSP::curfile" : '';
	print "<hr size=5></tt>";
	clear_buffer();
	die $e;
}

sub count_lines{
	my $text = shift;
        my @text = split /(\r?\n|\r)/, ".$text.";
        return scalar @text;
}

sub form_read_in{
	my($buffer, @data, $temp);
	binmode(STDIN);
	read(STDIN, $buffer, $ENV{CONTENT_LENGTH});
	if($buffer!~/\n/){
		$buffer .= '&' . $ENV{QUERY_STRING} unless $buffer;
		foreach $temp (split(/&|;/,$buffer)) {
			$temp =~ tr/+/ /;
			$temp =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
			$temp =~ s/\r//g;
			my @temp = split(/=/, $temp, 2);
			next unless @temp;
			if(exists $form{$temp[0]}){
				if(!ref $form{$temp[0]}){
					$form{$temp[0]} = [$form{$temp[0]}];
				}
				push @{$form{$temp[0]}}, $temp[1];
			}else{
				$form{$temp[0]}=$temp[1];
			}
		}
	}else{
		$buffer =~s/--(\r?\n)$/$1/;
		$buffer=~/(.*)(\r?\n)/;
		if($1 ne ''){
			my $bound = $1.$2;
			my @this=split(/\Q$bound\E/, $buffer);
			shift @this;
			foreach(@this){
				my($header, $body) = split(/\r?\n\r?\n/, $_, 2);
				my @head = split(/\r?\n/, $header);
				undef my %head;
				foreach(@head){
					my($name, $value) = split(/:/, $_, 2);
					if(lc $name eq 'content-disposition'){
						$head{name} = $1 if $value=~/name="?([^"]*)"?/;
						$head{filename} = $1 if $value=~/filename="?([^"]*)"?/;
					}elsif(lc $name eq 'content-type'){
						$head{"content-type"} = $1 if $value=~/\s*([^\s]*)\s*/;
					}
				}
				if($head{"content-type"} eq ''){
					# Regular Form Fields
					$body =~ s/\r?\n$//;
					if(exists $form{$head{name}}){
						if(!ref $form{$head{name}}){
							$form{$head{name}} = [$form{$head{name}}];
						}
						push @{$form{$head{name}}}, $body;
					}else{
						$form{$head{name}} = $body;
					}
				}else{
					$body =~ s/\r?\n$//;
					my %file;
					$file{type} = $head{"content-type"};
					$file{name} = $head{filename};
					$file{name} =~ s/.*[\\\/]//g;
					$file{name} =~ s/[^a-zA-Z0-9\-\_]/_/g;
					if($head{name}=~/\.([a-zA-Z0-9\_]+)/){
						$file{name} =~ s/\_(\Q$1\E)$/\.$1/
					}
					$file{data} = $body;
					my $name = $head{name};
					if(exists $form{$name}){
						if(!ref $form{$name}){
							$form{$name} = [$form{$name}];
						}
						push @{$form{$name}}, \%file;
					}else{
						$form{$name} = \%file;
					}
				}
			}
		}
	}
}

sub HSP_ECHO{ # rename print so we can delete it from error messages
	print @_;
}

sub htembed{
	my $file = shift;
	my $args = shift;
	local %form = () if ref $args eq 'HASH';
	if(ref $args eq 'HASH'){
		%form = %$args;
	}
	my $code = create_code($file); #get raw code, not a sub
	my $caller = $HSP::curfile;
	$HSP::curfile = $file;
	$code->(); # no eval! htembed is already within eval{}
	$HSP::curfile = $caller;
}

sub htsendheader{
	my %pairs = @_;
	chomp %pairs;
	foreach(keys %pairs){
		$HSP::header{lc $_} = $pairs{$_};
	}
}

sub htflush{
	clear_buffer();
}

package HSP::STDOUT;
sub TIEHANDLE{
	my $self = {};
	open(HANDLE, ">&STDOUT");
	$self->{HANDLE} = *HANDLE;
	bless $self, shift;
}
sub READ{undef}
sub READLINE{undef}
sub GETC{undef}
sub WRITE{undef}
sub PRINT{
	shift->{buffer} .= join($,, @_).$\;
}
sub PRINTF{
	shift->{buffer} .= sprintf($_[0], $_[1]);
}
sub CLOSE{
	shift->flush;
}
sub DESTROY{
	shift->flush;
}
sub flush{
	my $self = shift;
	*TIEDHANDLE = $self->{HANDLE};
	my $r = select TIEDHANDLE;
	$| = 1;
	local $\ = '';
	print $self->{buffer};
	$self->{buffer} = '';
	select $r;
}
sub prepend{
	my $self = shift;
	substr($self->{buffer},0,0) = join('', @_);
}

1;

